#include <stdlib.h>
#include <png.h>

#define PNG_OFFSET (127)

#define BOTTOM_LEVEL 5

#define BAYES 1
#define VISU  2
#define SURE  3
#define RNML  4
#define INML  5
#define SNML  6
#define MNML  7

#define HH 1
#define HL 2
#define LH 3

#define MAX(a,b) ((a)>(b)?(a):(b))
#define MIN(a,b) ((a)<=(b)?(a):(b))
#define for_each_coeff(row,col,range) for (row = (range).min_row; row < (range).max_row; row++) for (col = (range).min_col; col < (range).max_col; col++)
#define for_each_band(level,band,dim,minlevel,maxlevel) for (level = (minlevel); level <= (maxlevel); level++) for (band = 1; band <= (((level<1)||dim==1)?1:3); band++)
#define for_each_band_reverse(level,band,dim,minlevel,maxlevel) for (level = (maxlevel); level >= (minlevel); level--) for (band = 1; band <= (((level<1)||dim==1)?1:3); band++)

typedef struct signal_st {
  int dim;
  int len;
  int n;
  double *a;
  double **ap;
} Signal;

typedef struct indexed_value_st {
  int row;
  int col;
  double val;
} Ival;

typedef struct range_st {
  int min_row;
  int max_row;
  int min_col;
  int max_col;
  int n;
} Range;

// mdldenoise.c

void usage(char *);
Signal *read_signal(char *);
void write_sig(FILE *f, Signal *);
void write_signal(char *, Signal *);
Signal *copy_signal(Signal *);
double rmse(Signal *, Signal *);
double log10(double);
double psnr(Signal *, Signal *);
void free_signal(Signal *);
void wt(Signal *, int);
int log2dyad(int);
double square(double);
int comp_abs(const void *, const void *);
int comp_Ival(const void *, const void *);
double mad(Signal *);
double soft_threshold(double, double);
double hard_threshold(double, double);
int select_subband(Range *, Signal *, int, int);
double sort_range(Ival *, Signal *, Range *);
void simple_denoise(Signal *, int);
double codelength(Signal *, Ival ***, double ***, int **, int, int);
void iterative_denoise(Signal *, int);
void denoise(Signal *, int);

// pngopers.c

extern Signal *realloc_signal(Signal *, int);
extern Signal *alloc_signal(int, int);
extern int is_dyadic(int);
extern int is_png_fname(char *);
extern Signal *read_png(char *);
extern void write_png(char *, Signal *y);
