#include "mdldenoise.h"

Signal *realloc_signal(Signal *old, int len)
{
  Signal *new;
  int r;

  new = alloc_signal(old->dim, len);

  for (r = 0; r < (MIN(old->len, new->len)-1)*(old->dim-1)+1; r++)
    memcpy(new->ap[r], old->ap[r], sizeof **(new->ap)*MIN(old->len, new->len));

  free(old);
  return new;
}

Signal *alloc_signal(int dim, int len)
{
  Signal *y;
  int row;

  y = malloc(sizeof *y);

  y->dim = dim;
  y->len = len;
  if (dim == 1) // 1D
  {
    y->n = len;
    y->a = malloc(sizeof *(y->a)*y->n);
    y->ap = malloc(sizeof *(y->ap));
    y->ap[0] = y->a;
  }
  else         // 2D
  {
    y->n = len * len;
    y->a = malloc(sizeof *(y->a)*y->n);
    y->ap = malloc(sizeof *(y->ap)*y->len);
    for (row = 0; row < len; row++)
      y->ap[row] = y->a+row*len;
  }

  return y;
}

int is_dyadic(int x)
{
  while (!(x & 1)) x >>= 1;
  if (x==1) return 1; else return 0;
}

int is_png_fname(char *fname)
{
  if (strstr(fname, ".png") != NULL || strstr(fname, ".PNG") != NULL)
    return 1;
  else 
    return 0;
}

Signal *read_png(char *fname)
{
  Signal *y;
  int row, col;
  
  FILE *fp = fopen(fname, "rb");
  if (!fp)
    return NULL;

  png_structp png_ptr = png_create_read_struct
    (PNG_LIBPNG_VER_STRING, (png_voidp)NULL,
     NULL, NULL);
  if (!png_ptr)
    return NULL;

  png_infop info_ptr = png_create_info_struct(png_ptr);
  if (!info_ptr)
  {
    png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
    return NULL;
  }

  png_infop end_info = png_create_info_struct(png_ptr);
  if (!end_info)
  {
    png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
    return NULL;
  }

  if (setjmp(png_jmpbuf(png_ptr)))
  {
    png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
    fclose(fp);
    return NULL;
  }

  png_init_io(png_ptr, fp);

  png_read_info(png_ptr, info_ptr);

  // FORCE TO 8 BIT MODE

  int skip = 1;
  if (info_ptr->bit_depth < 8)
  {
    png_set_gray_1_2_4_to_8(png_ptr);
    skip = 3;
  }
  if (info_ptr->bit_depth == 16)
    png_set_strip_16(png_ptr);

  info_ptr->row_pointers = (png_bytep *) 
    malloc(sizeof(png_bytepp)*info_ptr->height);
  for (row = 0; row < info_ptr->height; row++)
    info_ptr->row_pointers[row] = malloc(info_ptr->rowbytes);
  
  png_read_image(png_ptr, info_ptr->row_pointers);

  png_read_end(png_ptr, end_info);

  if (info_ptr->height != info_ptr->width ||
      !is_dyadic(info_ptr->height))
  {
    fprintf(stderr, "size of image '%s' is %d x %d. not good.\n",
	    fname, (int)info_ptr->width, (int)info_ptr->height);
    exit(-1);
  }
  
  y = alloc_signal(2, info_ptr->height);

  for (row = 0; row < info_ptr->height; row++)
    for (col = 0; col < info_ptr->width; col++)
      y->ap[row][col] =
	(double)(info_ptr->row_pointers[row][col*skip])-PNG_OFFSET;

  png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);

  return y;
}

png_byte truncate_png_byte(double x)
{
  if (x < 0.5)
    return 0;
  else if (x > 245.5)
    return 255;
  else
    return (int)(x+0.5);
}

/* write a png file */
void write_png(char *fname, Signal *y)
{
   FILE *fp;
   png_structp png_ptr;
   png_infop info_ptr;
   int row, col;

   /* open the file */
   fp = fopen(fname, "wb");
   if (!fp)
      return;

   /* Create and initialize the png_struct with the desired error handler
      functions.  If you want to use the default stderr and longjump basis,
      you can supply NULL for the last three parameters.  We also check that
      the library version is compatible in case we are using dynamically
      linked libraries.
    */
   png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING,
				     (void *)NULL, NULL, NULL);

   if (!png_ptr)
   {
      fclose(fp);
      return;
   }

   info_ptr = png_create_info_struct(png_ptr);
   if (!info_ptr)
   {
      fclose(fp);
      png_destroy_write_struct(&png_ptr,  (png_infopp)NULL);
      return;
   }

   /* set error handling */
   if (setjmp(png_ptr->jmpbuf))
   {
      /* If we get here, we had a problem reading the file */
      fclose(fp);
      png_destroy_write_struct(&png_ptr,  (png_infopp)NULL);
      return;
   }

   /* set up the output control if you are using standard C streams */
   png_init_io(png_ptr, fp);


   /* if you are using replacement message functions, here you would call */
   //png_set_message_fn(png_ptr, (void *)msg_ptr, user_error_fn, user_warning_fn);
   /* where msg_ptr is a structure you want available to the callbacks */


   png_set_filter(png_ptr, 0, PNG_FILTER_NONE);

   png_set_IHDR(png_ptr, info_ptr, y->len, y->len,
                8, PNG_COLOR_TYPE_GRAY, PNG_INTERLACE_NONE,
                PNG_COMPRESSION_TYPE_DEFAULT,
                PNG_FILTER_TYPE_DEFAULT);

   /* write the file header information */
   png_write_info(png_ptr, info_ptr);

   info_ptr->row_pointers = (png_bytep *) 
     malloc(sizeof(png_bytepp)*info_ptr->height);
   for (row = 0; row < info_ptr->height; row++)
   {
     info_ptr->row_pointers[row] = malloc(info_ptr->rowbytes);
     for (col = 0; col < info_ptr->width; col++)
       info_ptr->row_pointers[row][col] = 
	 truncate_png_byte(y->ap[row][col]+PNG_OFFSET);
   }
  
   png_write_image(png_ptr, info_ptr->row_pointers);

   /* You can write optional chunks like tEXt, tIME at the end as well.
    * Note that if you wrote tEXt or zTXt chunks before the image, and
    * you aren't writing out more at the end, you have to set
    * info_ptr->num_text = 0 or they will be written out again.
    */

   /* write the rest of the file */
   png_write_end(png_ptr, info_ptr);

   /* if you malloced the palette, free it here */
   if (info_ptr->palette)
      free(info_ptr->palette);

   /* if you allocated any text comments, free them here */

   /* clean up after the write, and free any memory allocated */
   png_destroy_write_struct(&png_ptr,  (png_infopp)NULL);

   /* close the file */
   fclose(fp);

   /* that's it */
   return;
}
